/* 
Trivial example web server, using raw network sockets.

References:
    "Beej's Guide to Network Programming"
    http://beej.us/guide/bgnet/output/html/singlepage/bgnet.html

    "Creating a Basic Winsock Application", by Microsoft
    https://msdn.microsoft.com/en-us/library/windows/desktop/ms737629(v=vs.85).aspx

Dr. Orion Lawlor, lawlor@alaska.edu, 2016-03-24 (public domain)
*/
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <map>
#include <string>


// Socket headers:
#if defined(_WIN32) && ! defined(__CYGWIN__)
/* For windows systems:*/

//#  include <winsock.h> /* for SOCKET and others */
//#  pragma comment (lib, "wsock32.lib")  /* link with winsock library */

#  include <winsock2.h> /* new style, for getaddrinfo */
#  include <WS2tcpip.h> /* new style, for getaddrinfo */
#  pragma comment(lib, "Ws2_32.lib")
inline void close_socket_portable(int sock) { closesocket(sock); }

class initialize_winsock {
public:
	initialize_winsock() {
		WSADATA wsaData;
		WSAStartup(MAKEWORD(2,2), &wsaData);
	}
};
initialize_winsock initialize_winsock_now;

#else 
/* Non-windows (UNIX-y) systems: */
#  include <sys/types.h>
#  include <sys/socket.h>
#  include <netdb.h>
#  include <unistd.h>
inline void close_socket_portable(int sock) { close(sock); }
#endif

// Check if an error (negative return value) came back from this code.
#define SKT_CHECK(code) { int status=code; if (status<0) { perror(#code); std::cout<<"Error with code "<<#code<<" on line "<<__LINE__<<" of file "<<__FILE__<<"\n"; exit(1); } }

#include <sstream> // for ostringstream
#include <string.h> // for memset

// Read a newline-terminated string from this socket.
std::string recv_text_line(int socket) {
	std::string req="";
	char last_char=0;
	while (1) { 
		char c=0;
		SKT_CHECK(recv(socket,&c,1,0));
	
		if (c=='\n' || c=='\r') { // newline char
			if (last_char=='\r' && c=='\n') // full newline
				break;
		}
		else req+=c;
		last_char=c;
	}
	return req;
}



// Convert type to string
template <class T>
std::string to_string_portable(T t) {
	std::ostringstream s;
	s<<t;
	return s.str();
}


int main() {
	// Set up our network address:
	struct addrinfo hints;
	struct addrinfo *res;  // will point to the results

	memset(&hints, 0, sizeof hints); // make sure the struct is empty
	hints.ai_family = AF_UNSPEC;     // don't care IPv4 or IPv6
	hints.ai_socktype = SOCK_STREAM; // TCP stream sockets
	hints.ai_flags = AI_PASSIVE;     // fill in my IP for me
	
	// TCP port number to listen on:
	int server_port=8080;
	SKT_CHECK(getaddrinfo(NULL,to_string_portable(server_port).c_str(), &hints,&res));

	// Make a server socket:
	int serv=socket(res->ai_family, res->ai_socktype, res->ai_protocol);

	// Allow the server socket to be reopened within 2 minutes:
	int yes=1;
	SKT_CHECK(setsockopt(serv, SOL_SOCKET, SO_REUSEADDR, (const char *)&yes, sizeof(int)));
	
	// Give the server socket an address:
	SKT_CHECK(bind(serv,res->ai_addr,res->ai_addrlen)); // server
	freeaddrinfo(res);

	// Listen for client connections
	listen(serv, 10);
	
	std::cout<<"Listening at http://localhost:"<<server_port<<"\n";
	
	while (true) { // keep being a server forever

		std::cout<<"Waiting to accept() next client\n";   
		struct sockaddr_storage their_addr;
	    	socklen_t addr_size=sizeof(their_addr);
		int s=accept(serv,(struct sockaddr *)&their_addr,&addr_size);
		std::cout<<"Accepted client\n";	
	// Ideally you'd create a separate thread for this client here...

	/* Receive HTTP request headers, up to the newline */
		std::string req_URL="?";
		while (1) // (response=skt_recv_line(s))!="")
		{
			// Read a newline-terminated string:
			std::string req=recv_text_line(s);
			if (req=="") break;
		
			std::cout<<"	Header: "<<req<<"\n";
			if (req.substr(0,3)=="GET") {
				req_URL=req.substr(4);
				std::cout<<"GOT URL!  "<<req_URL<<"\n";
			}
		}

	/* Send the client back a web page. */
		std::string page_contents="This is web.  Sorta.\n\n";
		page_contents+="Your request URL was "+req_URL+"\n";
		std::string page_length=to_string_portable(page_contents.size());

		std::string response=
			"HTTP/1.1 200 OK\r\n"  // status code, e.g., 404 not found
			"Content-Length: "+page_length+"\r\n" // bytes in message
			"Content-Type: text/plain\r\n" // MIME type
			"\r\n" // blank line == end of HTTP request
			+page_contents; 

		SKT_CHECK(send(s,&response[0],response.size(),0));

		close_socket_portable(s);

	}

	return 0;
}





